; Minio
; by Adam Gashlin

    processor 6502

; Hardware registers
VRAMCN0 equ $2000
VRAMCN1 equ $2001
VRAMSTAT equ $2002
VRAMSCR equ $2005
VRAMADR equ $2006
VRAMIO  equ $2007
SPRDMA  equ $4014
JOY1    equ $4016

; Locations in VRAM space
PT      equ $0000
NT      equ $2000
AT      equ $23C0
BGPAL   equ $3F00
SPPAL   equ $3F10

SPRLOC  equ $0200   ; sprites in RAM

; constants
GRAVITY     equ $0028
FRICPOSACCEL    equ $0020
FRICNEGACCEL    equ $ffe0
POSACCEL    equ $0040
NEGACCEL    equ $ffc0
SPRHEIGHT   equ 8
SPRWIDTH    equ 7
JUMPACCEL   equ $fe00
INITIALOOMPH equ $15

; uninitialized data segment (zero page)
    SEG.U   data
    org $0000

TEMP    WORD
TEMP2   WORD
TEMP3   WORD
VRAMCN0_L BYTE
VRAMCN1_L BYTE
JOY1DATA BYTE

; velocity/position pairs
FIRSTVP
SPRVELY WORD
SPRPOSY WORD ; must be first for LASTY
SPRVELX WORD
SPRPOSX WORD
GNDVELX0 WORD
GNDPOSX0 WORD
GNDVELY0 WORD
GNDPOSY0 WORD
GNDVELX1 WORD
GNDPOSX1 WORD
GNDVELY1 WORD
GNDPOSY1 WORD
GNDVELX2 WORD
GNDPOSX2 WORD
GNDVELY2 WORD
GNDPOSY2 WORD
GNDVELX3 WORD
GNDPOSX3 WORD
GNDVELY3 WORD
GNDPOSY3 WORD
GNDVELX4 WORD
GNDPOSX4 WORD
GNDVELY4 WORD
GNDPOSY4 WORD
GNDVELX5 WORD
GNDPOSX5 WORD
GNDVELY5 WORD
GNDPOSY5 WORD
GNDVELX6 WORD
GNDPOSX6 WORD
GNDVELY6 WORD
GNDPOSY6 WORD
GNDVELX7 WORD
GNDPOSX7 WORD
GNDVELY7 WORD
GNDPOSY7 WORD
VPCOUNT equ 10
GNDCOUNT equ 8

GROUNDW0 BYTE    ; width
GROUNDW1 BYTE
GROUNDW2 BYTE
GROUNDW3 BYTE
GROUNDW4 BYTE
GROUNDW5 BYTE
GROUNDW6 BYTE
GROUNDW7 BYTE

; properties of the current ground
GROUNDW BYTE
GNDVELX WORD
GNDPOSX WORD
GNDVELY WORD
GNDPOSY WORD

RELVELX WORD
BESTRELVELY BYTE
SPRACCX WORD
; ACCY = GRAVITY if not on ground
ONGROUND BYTE
OOMPH   BYTE
INJUMP  BYTE
HOLDA   BYTE
HOLDA2  BYTE
HOLDDIR BYTE
FIRSTGNDVELY BYTE
LASTY   BYTE
NEXTLEV BYTE
LEVLOC  BYTE
LASTV

    SEG     text
    org $8000-16

; iNES header
; iNES magic
    BYTE $4e,$45,$53,$1a ; "NES,$1a
; Number of PRG-ROM blocks
    BYTE $01
; Number of CHR-ROM blocks
    BYTE $00
; ROM control bytes: Horizontal mirroring, no SRAM or trainer, Mapper #0
    BYTE $00, $00
; Padding
    BYTE $00, $00, $00, $00, $00, $00, $00, $00


; NMI
mainloop:
    jsr readcon
    sta JOY1DATA

    ; calculate relative velocity
    sec
    lda SPRVELX
    sbc GNDVELX
    sta RELVELX
    lda SPRVELX+1
    sbc GNDVELX+1
    sta RELVELX+1

    ; friction
    ldx #$00
    stx SPRACCX
    stx SPRACCX+1

    ; if not on ground no horizontal acceleration
    ldx ONGROUND
    bne xaccel
    lsr JOY1DATA
    lsr JOY1DATA
    jmp nolr
xaccel

    lda #[FRICPOSACCEL&$ff]
    ldy #[FRICPOSACCEL>>8]
    ldx RELVELX+1
    bne nonzeroxvel
    ldx HOLDDIR
    bne endoffric
    ldy GNDVELX
    sty SPRVELX
    ldy GNDVELX+1
    sty SPRVELX+1
    ldy #$00
    sty RELVELX
    beq endoffric
nonzeroxvel
    ldx RELVELX+1
    bmi posxvelfric
    lda #[FRICNEGACCEL&$ff]
    ldy #[FRICNEGACCEL>>8]
posxvelfric
    sta SPRACCX
    sty SPRACCX+1
endoffric

    ldx #$00
    txa
    tay

    lsr JOY1DATA
    bcc noright
    ldx #$05    ; velocity limit
    cpx RELVELX+1 ; 5-vel
    bmi noright
    beq noright
    ldy #[POSACCEL>>8]
    lda #[POSACCEL&$ff]
noright
    lsr JOY1DATA
    bcc noleft
    ldx #$fb    ; velocity limit
    cpx RELVELX+1 ; -5-vel
    bpl noleft
    ldy #[NEGACCEL>>8]
    lda #[NEGACCEL&$ff]
noleft
    stx HOLDDIR
    clc
    adc SPRACCX
    sta SPRACCX
    tya
    adc SPRACCX+1
    sta SPRACCX+1
nolr
    lsr JOY1DATA
    bcc nodown
nodown
    lsr JOY1DATA
    bcc noup
noup
    lsr JOY1DATA
    bcc nostart
nostart
    lsr JOY1DATA
    bcc noselect
noselect
    lsr JOY1DATA
    bcc noB
noB

    lda #$00
    tay
    lsr JOY1DATA
    bcc noA
    ldy #1
    ldx ONGROUND
    beq notstartjump
    ldx HOLDA
    bne notstartjump
    ldx #INITIALOOMPH
    stx INJUMP
    stx OOMPH
    ldx GNDVELY+1
    stx FIRSTGNDVELY
notstartjump
    ldx INJUMP
    beq noA
    ldx OOMPH
    beq noA
    dex
    stx OOMPH
    sta ONGROUND
    lda #[JUMPACCEL&$ff]
    sta SPRVELY
    clc
    lda #[JUMPACCEL>>8]
    adc FIRSTGNDVELY
    sta SPRVELY+1
    lda #1
noA
    sta INJUMP
    lda HOLDA2
    sta HOLDA
    sty HOLDA2

    ; perform acceleration

    lda SPRACCX
    clc
    adc SPRVELX
    sta SPRVELX
    lda SPRACCX+1
    adc SPRVELX+1
    sta SPRVELX+1

    ; perform movement

    ldy #[VPCOUNT*4]-1

moveloop
    dey
    dey
    dey
    clc
    lda FIRSTVP+2,y
    adc FIRSTVP+0,y
    sta FIRSTVP+2,y
    lda FIRSTVP+3,y
    sta LASTY
    adc FIRSTVP+1,y
    sta FIRSTVP+3,y
    dey
    bpl moveloop

    ; Check for end of level

    lda SPRPOSY+1
    and #$f0
    eor #$f0
    beq jreboot
    lda SPRPOSY+1
    and #$fc
    bne noreboot
    lda NEXTLEV
    sta LEVLOC
jreboot
    jmp reboot
noreboot

    ; process each ground unit
    ldx #0
    stx ONGROUND
    dex
    stx BESTRELVELY
    ldy #GNDCOUNT

    lda #[SPRLOC>>8]
    sta TEMP+1
    lda #$04
    sta TEMP

groundloop
    dey
    bmi donewithground
    lda GROUNDW0,y
    bne notgroundloop
    jmp groundloop
donewithground
    jmp endofgroundloop
notgroundloop
    sta TEMP3+1
    tya
    pha
    asl
    asl
    asl
    tay

    ; collision w/ ground

    ; check x coord
    lda SPRPOSX+1
    sec
    sbc GNDPOSX0+1,y ; A-GNDPOSX
    cmp TEMP3+1
    bpl nohitground ; if to the right
    eor #$ff
    cmp #(SPRWIDTH-1)
    bpl nohitground ; if to the left

    ; check y coord and velocity
    lda GNDPOSY0,y
    sec
    sbc GNDVELY0,y
    lda GNDPOSY0+1,y
    tax
    sbc GNDVELY0+1,y
    sec
    sbc #SPRHEIGHT
    cmp LASTY
    bmi nohitground ; if below ground before
    txa
    sec
    sbc #SPRHEIGHT
    cmp SPRPOSY+1
    beq ongndnow
    bpl nohitground ; if above ground now

ongndnow
    sec
    lda SPRVELY
    sbc GNDVELY0,y
    lda SPRVELY+1
    sbc GNDVELY0+1,y
    bmi nohitground ; if going up relative to ground
    cmp BESTRELVELY
    bmi nohitground ; if slower than competing grounds

    ; collision detected
    sta BESTRELVELY

    ldx #1
    stx ONGROUND
    lda GNDVELX0,y
    sta GNDVELX
    lda GNDVELX0+1,y
    sta GNDVELX+1
    lda GNDVELY0,y
    sta GNDVELY
    lda GNDVELY0+1,y
    sta GNDVELY+1

    lda GNDPOSX0,y
    sta GNDPOSX
    lda GNDPOSX0+1,y
    sta GNDPOSX+1
    lda GNDPOSY0,y
    sta GNDPOSY
    lda GNDPOSY0+1,y
    sta GNDPOSY+1
    lda TEMP3+1
    sta GROUNDW

    ; onground
    lda GNDVELY+1
    sta SPRVELY+1
    lda GNDVELY
    sta SPRVELY
    lda GNDPOSY
    sta SPRPOSY
    lda GNDPOSY+1
    sec
    sbc #SPRHEIGHT
    sta SPRPOSY+1

nohitground

    ; ground sprites
    lda TEMP3+1
    pha
    lda GNDPOSX0+1,y
    sta TEMP3
drawgroundloop
    lda GNDPOSY0+1,y
    ldx #$00
    sta (TEMP),x ; y coord
    inc TEMP
    lda #$01
    sta (TEMP),x  ; tile #
    inc TEMP
    lda #$00
    sta (TEMP),x ; attrib
    inc TEMP
    lda TEMP3
    sta (TEMP),x ; x coord
    inc TEMP
    clc
    adc #$08
    sta TEMP3
    pla
    sec
    sbc #$08
    pha
    bne drawgroundloop
    pla

    pla
    tay
    jmp groundloop
endofgroundloop

    ; gravity
    lda ONGROUND
    bne nograv
    lda SPRVELY
    clc
    adc #[GRAVITY&$ff]
    sta SPRVELY
    lda SPRVELY+1
    adc #[GRAVITY>>8]
    sta SPRVELY+1
nograv

    ; update primary sprite

    lda SPRPOSY+1
    sta SPRLOC    ; y coord
    lda SPRPOSX+1
    sta SPRLOC+3  ; x coord
    lda #$01
    sta SPRLOC+1  ; tile #
    lda #$00
    sta SPRLOC+2  ; attrib

    ; DMA sprites
    lda #[SPRLOC>>8]
    sta SPRDMA

    rti

; Init routine based on Chris Covell's

start:
   cld         ;Clear decimal flag
   sei         ;Disable interrupts

   ldx #$00
   stx LEVLOC

reboot:
   ldx #$00
   stx VRAMCN0
   stx VRAMCN1
   dex
   txs         ;Top of stack at $1FF

   ldx LEVLOC

;Clear (all of) the NES' WRAM.
   ldy #$07
   sty TEMP+1
   ldy #$00
   sty TEMP
   lda #$00

Clear   sta (TEMP),y    ;Clear $100 bytes
   dey
   bne Clear

   dec TEMP+1    ;Decrement "banks" left counter
   bpl Clear     ;Do next if >= 0

   stx LEVLOC

   lda   #$20
   sta   VRAMADR
   lda   #$00
   sta   VRAMADR

WaitV   ldx VRAMSTAT
   bpl WaitV

WaitV2   ldx VRAMSTAT
   bpl WaitV2

   ldx   #$00
   ldy   #$10
ClearPPU sta VRAMIO  ;Clear the PPU space.  REALLY IMPORTANT for a real NES!
   dex
   bne   ClearPPU
   dey
   bne   ClearPPU

   sta SPRDMA

    ; fill pal
    lda #[BGPAL >> 8]
    sta VRAMADR
    lda #[BGPAL & $FF]
    sta VRAMADR

    ldx #$00
palloop
    stx VRAMIO
    inx
    cpx #$20
    bne palloop

    ; fill pattern table
    lda #$00
    sta VRAMADR
    sta VRAMADR

    ldx #$00
    lda #$00
    sta TEMP
    sta TEMP2

ptloop2
    ; this repetition is not done lightly, it 3 bytes shorter than the
    ; complex version

    lda TEMP
    ldy #$08
ptloop
    sta VRAMIO
    dey
    bne ptloop

    lda TEMP2
    ldy #$08
ptloop1
    sta VRAMIO
    dey
    bne ptloop1

    lda TEMP
    eor #$ff
    php
    sta TEMP
    lda TEMP2
    plp
    bne pt_nocarry
    eor #$ff
pt_nocarry
    sta TEMP2

    inx
    cpx #$04
    bne ptloop2

    ; Reset VRAM address reg
    lda #$00
    sta VRAMADR
    sta VRAMADR

    ; Load level

    ldx LEVLOC
levelloop
    inx
    ldy leveldata-1,x
    beq leveldone
    lda leveldata,x
    sta 0,y
    inx
    jmp levelloop
leveldone
    stx NEXTLEV

    ; Set PPU attributes
    lda VRAMCN0_L
    ora #$80
    sta VRAMCN0_L
    sta VRAMCN0
    lda VRAMCN1_L
    ora #$1e    ; sprites, bg visible, no clipping
    sta VRAMCN1_L
    sta VRAMCN1

deadend
    jmp deadend

; return controller data in A
readcon subroutine
    ldx #$08
    lda #$01
    sta JOY1
    stx JOY1
.loop
    lda JOY1
    ror
    rol TEMP
    dex
    bne .loop

    lda TEMP

    rts

; Two byte address:value pair
; Address of 0 ends level
leveldata:
    ; level 1
    BYTE GNDVELX0,$80
    BYTE GNDPOSY0+1,$30
    BYTE GROUNDW0,$40

    BYTE GNDVELX1+1,$ff
    BYTE GNDVELX1,$80
    BYTE GNDPOSY1+1,$60
    BYTE GROUNDW1,$40

    BYTE GNDVELX2,$80
    BYTE GNDPOSX2+1,$30
    BYTE GNDPOSY2+1,$90
    BYTE GROUNDW2,$40

    BYTE SPRPOSY+1,$88
    BYTE SPRPOSX+1,$50
    BYTE 0

; level 2
    BYTE GNDVELX0,$80
    BYTE GNDPOSY0+1,$30
    BYTE GROUNDW0,$08

    BYTE GNDVELX1+1,$ff
    BYTE GNDVELX1,$80
    BYTE GNDPOSY1+1,$60
    BYTE GROUNDW1,$10

    BYTE GNDVELX2,$80
    BYTE GNDPOSX2+1,$30
    BYTE GNDPOSY2+1,$90
    BYTE GROUNDW2,$40

    BYTE SPRPOSY+1,$88
    BYTE SPRPOSX+1,$50
    BYTE 0

; level 3
    BYTE GNDVELY0,$80
    BYTE GNDPOSX0+1,$20
    BYTE GNDPOSY0+1,$50
    BYTE GROUNDW0,$40

    BYTE GNDVELY1+1,$ff
    BYTE GNDPOSX1+1,$b0
    BYTE GNDPOSY1+1,$50
    BYTE GROUNDW1,$40

    BYTE SPRPOSY+1,$48
    BYTE SPRPOSX+1,$40
    BYTE 0

; level 4
    BYTE GNDVELY0,$80
    BYTE GNDPOSX0+1,$20
    BYTE GNDPOSY0+1,$50
    BYTE GROUNDW0,$40

    BYTE GNDVELY1+1,$ff
    BYTE GNDPOSX1+1,$b0
    BYTE GNDPOSY1+1,$50
    BYTE GROUNDW1,$10

    BYTE SPRPOSY+1,$48
    BYTE SPRPOSX+1,$40
    BYTE 0

; level 5
    BYTE GNDVELY0,$80
    BYTE GNDPOSX0+1,$20
    BYTE GNDPOSY0+1,$50
    BYTE GROUNDW0,$20

    BYTE GNDVELY1+1,$ff
    BYTE GNDPOSX1+1,$b0
    BYTE GNDPOSY1+1,$50
    BYTE GROUNDW1,$10

    BYTE SPRPOSY+1,$48
    BYTE SPRPOSX+1,$24
    BYTE 0

 org $bffa,0

 WORD mainloop ; NMI
 WORD start ; RESET
 WORD start ; IRQ/BRK

; ocremix.org ROCKS!
